#include "includes.h"

DECLARE_WAIT_QUEUE_HEAD(qWaitforDeviceReady);

NTSTATUS IFDGetCardStatus( PDEVICE_EXTENSION pDeviceExtension )
{
	NTSTATUS 					NTStatus = STATUS_SUCCESS;
	PSCARD_READER_GETCARDSTATUS_PARAMS		pScardStatus= NULL;
	PSCARD_CCID_COMMAND_PARAM			pCCIDParam=NULL;
	u_char						*pResBuffer=NULL;
	u_char						byIndex = 0;
	u_char						byStatus = 0;
	PDWORD						pdwDataLen;

	do
	{
		pdwDataLen = (PDWORD) kmalloc( (sizeof(DWORD) * 1), GFP_KERNEL );
		if( NULL == pdwDataLen )
		{
			NTStatus = STATUS_INSUFFICIENT_RESOURCES;
			break;
		}
		memset( pdwDataLen, 0, (sizeof(DWORD)*1));
		
		pCCIDParam =  (PSCARD_CCID_COMMAND_PARAM)
			kmalloc (sizeof(SCARD_CCID_COMMAND_PARAM), 
			GFP_KERNEL );
		if (NULL == pCCIDParam)
		{
			NTStatus = STATUS_INSUFFICIENT_RESOURCES;
			break;
		}
		memset(pCCIDParam, 0, sizeof(SCARD_CCID_COMMAND_PARAM));


		pResBuffer =  (u_char *) kmalloc ( 
			(sizeof(u_char) * MAX_CCID_OUTBUFF_SIZE), 
			GFP_KERNEL );
	 	if (NULL == pResBuffer)
		{
			NTStatus = STATUS_INSUFFICIENT_RESOURCES;
			break;
		}
 		memset(pResBuffer, 0, MAX_CCID_OUTBUFF_SIZE);

 		pScardStatus =  (SCARD_READER_GETCARDSTATUS_PARAMS *)
			kmalloc (sizeof(SCARD_READER_GETCARDSTATUS_PARAMS), 
			GFP_KERNEL );
 	 	if (NULL == pScardStatus)
		{
			NTStatus = STATUS_INSUFFICIENT_RESOURCES;
			break;
		}
 		memset(pScardStatus, 0, sizeof(SCARD_READER_GETCARDSTATUS_PARAMS));

		if( TRUE == pDeviceExtension->bBootRomPresent )
		{
			DebugLog("IFDGetCardStatus::Device is in Bootrom Mode \n" );
			break;
		}

		// Build Command header
		pCCIDParam->Header.byMessageType	= PC_TO_RDR_GETSLOTSTATUS;
		pCCIDParam->Header.dwLength		= 0;
		pCCIDParam->Header.bySlot		= MASTER_SLOT_ID;
		pCCIDParam->Header.bySequence		= 0x01;

		// RFU bytes
		for (byIndex = 0; byIndex < MAX_RFU_BYTES; byIndex++)
		{
			pCCIDParam->Header.abyMessageSpecific [byIndex] = 0x00;
		}

		// Data Length + Command Header Length
		pDeviceExtension->ulDataLength = \
				pCCIDParam->Header.dwLength + CCID_HEADER_SIZE;

		DebugLog("GETCARDSTATUS %ld, \n", pDeviceExtension->ulDataLength );

		// Write the command
		NTStatus = PcdmWrite ( 
							pDeviceExtension, 
							(u_char *)pCCIDParam, 
							pDeviceExtension->ulDataLength
							);
		if (!NT_SUCCESS (NTStatus))
		{
			break;
		}

		// Read the response
		NTStatus = PcdmRead ( pDeviceExtension, pResBuffer, pdwDataLen );
		if (!NT_SUCCESS (NTStatus))
		{	
			break;
		}

		// Check the Response Type and Sequence Number

		if ((RDR_TO_PC_SLOTSTATUS == pResBuffer[MESSAGE_TYPE_OFFSET]) && \
			(0x01 == pResBuffer[SEQUENCE_OFFSET]))
		{
			pScardStatus->byStatus	= pResBuffer[STATUS_REG_OFFSET];
			pScardStatus->byError	= pResBuffer[ERROR_REG_OFFSET];

			if (CCID_ICC_ABSENT == pScardStatus->byStatus)
			{
				pScardStatus->byCardStatus = READER_STS_CARD_ABSENT;
				pDeviceExtension->byCardStatus = READER_STS_CARD_ABSENT;
				DebugLog ("<1> PcdmHdlr::Card ABSENT !!!(%d)\n", READER_STS_CARD_ABSENT);
			}
			else
			{
				if ((CCID_ICC_PRESENT == pScardStatus->byStatus) || \
					(CCID_ICC_ACTIVE == pScardStatus->byStatus))
				{
					pScardStatus->byCardStatus = READER_STS_CARD_PRESENT;
					pDeviceExtension->byCardStatus = READER_STS_CARD_PRESENT;
					DebugLog ("<1> PcdmHdlr::Card PRESENT !!! (%d)\n",READER_STS_CARD_PRESENT);
				}
				else
				{
					NTStatus = STATUS_UNSUCCESSFUL;
				}
			}
		}

		byStatus = pScardStatus->byCardStatus;
		pDeviceExtension->cardStsChange = FALSE;
		if(pDeviceExtension->prevCardStatus != byStatus)
		{
			pDeviceExtension->cardStsChange = TRUE;
		}
		pDeviceExtension->prevCardStatus = byStatus;
		pDeviceExtension->bySequence++;

	} while( FALSE );

	if( pdwDataLen) 
	{
		kfree( pdwDataLen );
		pdwDataLen = NULL;
	}

	if (pCCIDParam)
	{
		kfree (pCCIDParam);
		pCCIDParam = NULL;
	}

	if (pResBuffer)
	{
		kfree(pResBuffer);
		pResBuffer = NULL;
	}

	if (pScardStatus)
	{
		kfree (pScardStatus);
		pScardStatus = NULL;
	}

	if( STATUS_SUCCESS == NTStatus )
	{
		return byStatus;
	}
	else
	{
		return NTStatus;
	}
}

BOOLEAN
PcmciaIrqServiceRoutine (
        PDEVICE_EXTENSION       pDeviceExtension
        )
/*++

PcmciaIrqServiceRoutine:
        When interrupted the call is passed to the DPC routine immediately
        and the IRQ is reported as serviced

Arguments:
        Interrupt                       - Interrupt object related to the interrupt
        pDeviceExtension        - Context of call

Return Value:
        TRUE
        FALSE

--*/
{
	PPCDM_REGISTERS 	IoBase;
	UCHAR				Status;
	BOOLEAN				bFlag = FALSE;
	UCHAR   			DummyReg6;
		
	//DebugLog("<1> PcmciaIrqServiceRoutine Entry \n" );
	IoBase = (PPCDM_REGISTERS) pDeviceExtension->IoBase;

	do
	{
		Status = inb( (unsigned) &IoBase->CmdStatusReg );


		// Read ATA Command/Status register to clear the interrupt
		if( TRUE == pDeviceExtension->bDwnComplete )
		{
				Status = inb( (unsigned) &IoBase->CmdStatusReg);
				pDeviceExtension->bDwnComplete = FALSE;
		}
		if( TRUE == pDeviceExtension->bBootRomPresent )
		{
			DebugLog("<1> The device is in boot rom mode. \n");
			break;
		}
		
		Status = inb( (unsigned) &IoBase->CmdStatusReg );
		if (TRUE == (Status & STS_ERR))
		{
			bFlag = FALSE;	
			break;
		}
	
		DummyReg6 = inb ( (unsigned) &IoBase->DummyReg6 );

		if ((0x01 != DummyReg6) && (0x02 != DummyReg6))
		{
			bFlag = FALSE;	
			break;					
		}

		bFlag = TRUE;


		if (0x02 == ( DummyReg6 & 0x02))  //Data Ready
		{
			// We only clear the 2nd bit since we may also have
			// got card status change
			DummyReg6 &= 0xFD;

			outb( DummyReg6 , (unsigned) &IoBase->DummyReg6 );
			DebugLog("<1> SCR241 Pcmcia Module!Signal event.. ");
			if( FALSE == pDeviceExtension->bSignalEvent )
			{
				pDeviceExtension->bSignalEvent = TRUE;
				if( TRUE == pDeviceExtension->bIsEventSet )
				{
					DebugLog("<1> Proper Setting of event \n");
					pDeviceExtension->bIsEventSet = FALSE;
					wake_up_interruptible_sync( &pDeviceExtension->tqsEventWait ); 
				}
			}												
		}

		if (0x01 == ( DummyReg6 & 0x01))  //card status change
		{
			DebugLog(" <1> Card Status Change. Schedule Event\n" );
			// we only clear the Cardstatus change event bit.
			DummyReg6 &= 0xFE;
			if( FALSE == pDeviceExtension->bIsPolledMode )
			{
				outb( DummyReg6, (unsigned) &IoBase->DummyReg6 );
				pDeviceExtension->bIsQueueSet = TRUE;
				schedule_task( &pDeviceExtension->tqsCardStatusCheck ); 
			}
		}					
	} while (FALSE);
	return bFlag;
}

// End of file
