#ifndef _PCSC_H_
#define _PCSC_H_

#include "includes.h"

#if !defined( STATUS_DEVICE_REMOVED )
#define STATUS_DEVICE_REMOVED	STATUS_UNSUCCESSFUL
#endif

// Device state flags
#define DEVICE_FLAG_INITIALIZING	0x00000001
#define DEVICE_FLAG_STOPPED		0x00000002
#define DEVICE_FLAG_QUERY_STOPPED	0x00000004
#define DEVICE_FLAG_CONNECTED		0x00000008
#define DEVICE_FLAG_REMOVED		0x00000010
#define DEVICE_FLAG_SURPRISE_REMOVED	0x00000020
#define DEVICE_FLAG_CLAIMED		0x00000040
#define DEVICE_FLAG_QUEUE_LOCKED	0x00000080
#define DEVICE_FLAG_QUEUE_FLUSHING	0x00000100
#define DEVICE_FLAG_DELETED		0x00000200

#define SMARTCARD_POOL_TAG 'DICC'

// Server registry entries
#define MAX_DEVICES_SUPPORTED		0x04
#define MAX_REG_ENTRIES			0x0D

// bRequest Value for DFU Interface
#define DFU_DETACH			0x00

#define LOCKED_STATE			1
#define UNLOCKED_STATE			0

// General Scard Definitions
#define ATR_SIZE			0x40
#define MAX_IFSD			252
#define STC2_CLK_FREQ			3680

#define SMARTCARD_DEVICE_TYPE 		0x32
#define BULKIN_ENDPOINT_SIZE  		64

//The following define is used to identify the current personality of 
//the device , whether PINPAD supported or not.
#define PIN_PRESENT  0x00 // pinpad not present

#define  NONSIGNALEDSTATE 0
#define  SIGNALEDSTATE 1

#define MAX_DATA_SIZE	4096
#define PCDM_REV_A      0x00
#define PCDM_REV_C      0x01

// This structure is used in the transfer of 
// data from the user area to the kernel area
// for read/write operations.
typedef struct _TXRX_DATA
{
	UCHAR ucBuffer[MAX_DATA_SIZE]; //joe
	ULONG ulDataLen;
}TXRX_DATA, *PTXRX_DATA;
 
typedef struct _DISPATCHER_HEADER 
{
    UCHAR Type;
    UCHAR Absolute;
    UCHAR Size;
    UCHAR Inserted;
    LONG SignalState;
//    LIST_ENTRY WaitListHead;
} DISPATCHER_HEADER;


typedef struct _KEVENT 
{
	DISPATCHER_HEADER Header;
} KEVENT, *PKEVENT; //, *RESTRICTED_POINTER PRKEVENT;

                                                                                                                                               
 
/*********************************************
	This structure is used to handle
	all the IOCTL requests from the 
	application
**********************************************/
typedef struct _IOCTL_DATA
{
	//UCHAR	ucTxBuffer[MAX_DATA_SIZE];
	UCHAR	ucTxBuffer[2048];
	DWORD	dwTxLength;
	//UCHAR	ucRxBuffer[MAX_DATA_SIZE];
	UCHAR	ucRxBuffer[2048];
	DWORD	dwRxLength;
}IOCTL_DATA, *PIOCTL_DATA;
 
typedef struct _DEVICE_EXTENSION *PDEVICE_EXTENSION;

/* a structure to store all information we need
   for our thread */
struct _kthread_struct
{
        /* A Back pointer to the device extension */
	PDEVICE_EXTENSION pDeviceExtension;
        /* private data */
        /* Linux task structure of thread */
        struct task_struct *thread;
        /* Task queue need to launch thread */
        struct tq_struct tq;
        /* function to be started as thread */
        void (*function) (struct _kthread_struct *kthread);
        /* semaphore needed on start and creation of thread. */
        struct semaphore startstop_sem;
        /* public data */
        /* queue thread is waiting on. Gets initialized by
           init_kthread, can be used by thread itself.
        */
        wait_queue_head_t queue;
        /* flag to tell thread whether to die or not.
           When the thread receives a signal, it must check
           the value of terminate and call exit_kthread and terminate
           if set.
        */
        int terminate;
        /* additional data to pass to kernel thread */
        void *arg;
};
// forward definition for the polled mode cardtracking thread structure
typedef struct _kthread_struct kthread_t;

typedef struct _DEVICE_EXTENSION
{
	// The current number of io-requests
	LONG						IoCount;
	// Irp used for all Io operations
	//PIRP				CurrentIrp;
	// Flag used to identify device status
	ULONG						DeviceFlags;
	// To identify whether reader active state
	BOOLEAN						ReaderOpen;
	// Power manager related variables
//	DEVICE_POWER_STATE				CurrentDevicePowerState;
	// the following fields hold info abt the reader
	BYTE						byReaderMode;
	// Store CCID Current Level param
	BYTE						byLevel;
	// flag for geting the current status from the bulk-in/out endpoint
	BOOLEAN						bFirstInsertion;
	// CCID Xfer length
	ULONG						ulDataLength;
	// CCID Sequence number
	BYTE						bySequence;
	// pending notification IRP and related flags
	BOOL						bNotifyIrpPending;
	// card removal identifier for pinpad reader
	BOOL						bMute;
	// For Port number
	int							nPortNumber;
	PUCHAR  					pIOBuffer;
	ULONG   					ulOutBufferSize;
	UCHAR   					byAbortSequence;
	UCHAR 						ucIntdata[2];
	// Bus drivers set the appropriate values in this structure in response
	// to an IRP_MN_QUERY_CAPABILITIES IRP. Function and filter drivers might
	// alter the capabilities set by the bus driver.
 
	// SPR Pinpad related variables
	BOOL						bSecureRead;
	BOOL						bSecureReadPending;
	BOOL						bModify;
	BYTE						byCardID;
	BYTE						byCardVal;
	// Cancel Thread variables
	BOOL						bThreadStarted;
	BOOL						bThreadProgress;
	BOOL						bStopThread;
	//Flag to decide only read mode.
	BOOLEAN 					bOnlyRead;
	// Used to signal that the reader is able to process reqeusts
	KEVENT						ReaderStarted;
	// Used to signal the the reader has been closed
	KEVENT						ReaderClosed;
	//Serial related
	KEVENT  					SerialCloseDone;
	// This structure is used to handle the I/O of IOCTLs
	IOCTL_DATA					IoctlData;
	// These booleans are used to set trigger a cold reset
	// from the daemon
	BOOL						bSimulateReset;
	
	int							nInstanceCount;
	
	// Pointer to structure DEV_LINK_T - pcmcia specific
	dev_link_t 					*DeviceLink;
	
	// Pointer to array containing the device name and the instance.
	PUCHAR						pucDeviceName;
	
	// Major number of the device thats allocated by the OS
	int							nMajorNumber;
	
	// Base Address of PCDM Register
	u_long						IoBase;
	
	// Base of mapped attribute memory	
	caddr_t 					am_base;	
	
	BYTE						byCardStatus;
	
	spinlock_t					cmd_lock;
	spinlock_t					rdwr_lock;
	spinlock_t					ioctl_lock;
	struct tq_struct			tqueue;
	
	wait_queue_head_t			wait;
	wait_queue_head_t			delay;
		
	u_char						irqFlag;
	u_char  					inIOCTL;
	u_char						FwDwnLdInProgress;
	dev_node_t 					DeviceNode;
	struct bus_operations		*bus;
	int							nAssignedIRQ;
	int  						nStop;
	ULONG						tempCount;
	u_char 						prevCardStatus;
	u_char						cardStsChange;
	BOOL						bWaitingforCardUpdate;
	BOOL						bNoIRQ;
	BOOL						bSignalEvent;
	BOOL						bIsIrqSet;
	/* structure used for polled mode card tracking */
	//struct _kthread_struct				CardTracker;
	kthread_t					CardTracker;
	// Flag to check/set forced polled mode.
	BOOL						bIsPolledMode;
	// To store the Revision of the Reader Plugged-in
	// Rev-A = 0. Rev-C = 1
	BYTE						byPcdmRev;
	// DFU
	BOOL						bBootRomPresent;
	BOOLEAN						bDwnComplete;
	BOOLEAN						b256WordDfu;
	PUCHAR						pucFwBuffer;
	ULONG						ulOffset;
	//Boolean to signal read-ready event
	BOOLEAN						bIsEventSet;
	struct tq_struct			tqsCardStatusCheck;
	wait_queue_head_t			tqsEventWait;
	BOOLEAN						bIsQueueSet;
	BOOLEAN						bOrgIrqStatus;
	BOOLEAN						bIsSystemResume;
} DEVICE_EXTENSION;

typedef struct _READER_EXTENSION
{
	// Driver device extension
	PDEVICE_EXTENSION				DeviceExtension;
	// miscellaneous
	ULONG						ReaderPowerState;
	//notused
	UCHAR						FirmwareMajor;
	UCHAR						FirmwareMinor;
	// For NTT Communication ...
	BOOLEAN						bLockState;
	//This boolean is used in serialioctl
	BOOLEAN         				bInternalDeviceIoControl;
	//Reader info details
	SCARD_READER_GETINFO_PARAMS  			Readerinfo;
	//***********************************************
	//Serial Related 
	//***********************************************
	//Follwing structure is used to store the basic setting values
	//aquired from serialport , prior to our initialization.
	//These settings are restored to the serial driver before close the serial port
	//SERIAL_PORT_CONFIG				SerialPortConfig;
	//For calls routed to reader,Timeout is implemented
	//to receive the start of response frame
	//the following boolean is used to decide forthcoming 
	//serial reponse timeout ( whether infinite or NOT )
	BOOLEAN						bWaitInifinitelyforResponse;
	//Serial frame recieve buffer used in readcallback
	UCHAR           				SerialBuffer[ MAX_DATA_SIZE ];
	UCHAR							FrametobeSent[MAX_DATA_SIZE ];
	//Length of frame built ( Frametobesent)
	WORD            				wFramelength;
	//Following booleans are used in scardupdatecardstate* functions
	//during clearing the reader extension Frame buffers
	BOOLEAN						bFrameTobeSentValid;
	BOOLEAN						bToClearFrameTobeSent;
	BOOLEAN						bFrameRecievedValid;
	BOOLEAN						bToClearRecievedFrame;
	//This used in serialtransmitrecieve and serialreadcallback for
	//handling unexpected ( non-needed) non dynamic frames.
	BOOLEAN         				bExpectFrame;
	//The following boolean is used in serialbuildframe and IFDctrlabort to decide
	//the startbyte  to be set.
	BOOLEAN						bAbortCommandIssued;
	//This flag is used in Serail frame striping function to check the validity of
	//SerialBuffer .
	BOOLEAN						bFrameReady;
	//This variable is used in readcallback. this hold the total bytes read.
	LONG            				lReadcount;
	//Following boolean are used to decide ONLYWRITE are needed ( EMV timed mode)
	BOOLEAN         				bOnlyWrite;
	ULONG						ulToCheck;
	//This is used in serial setcallback function
	ULONG						EventMask;
	BOOLEAN						bSerialDriverClosed;
	//This variable is used to hold the current COM port, in which the device is connected
	UCHAR							ComportNumber;
    //After the device/driver goes to suspend/hibernate state,
    //we block all functions routed through IFtoSerial ,
    //until the completion of devicepowercompletion routine
    //with exception to IFDcardstatus call from devicepower completion
    //routine.We usethe following boolean for that purpose in
    //Devicepowercompletion routine and IFDtoSerial
    BOOLEAN						bExceptionalCallfromDevicePwrCompRoutine;

    //Following event is used to signal that the frame reception
    //has started.(This is not used to signal start of Dynamic frame)
    //This variable is used in SerialTransmitRecieve and SerialReadcallback
    KEVENT						FrameStarted;
                                                                                                                                          
    //This event is used to signal the driver thread, from readcallback(system thread)
    //that a frame is completely read.
    KEVENT						FrameReadCompleted;

	// This flag is used to synchronize between dynamic frame handling and
	// command response data handling's update card state
	BOOLEAN						bUpdateCardState;

	//Dynamic Serial frame recieve buffer used in readcallback
	UCHAR           				DynamicSerialBuffer[ MAX_DATA_SIZE ];
	// This event is used to signal to the serialread thread to start waiting/reading the 
	// next set of frame from the serial port.
	KEVENT						FrameChkCompleted;

	// This variable is used to pass the IOCTL sent from the application
	DWORD						dwIoctlCode;
	
	// This flag is used to check if PPS is required for the current card.
	BOOLEAN						bIsPPSRequired;
	
	// This flag is used to check if PPS has already been done
	BOOLEAN						bIsPPSDone;
	
	// This flag is used to see if this is the first transmit being made
	// after ATR retrieval.
	BOOLEAN						bIsFirstTransmit;

	// This flag is used to check if PTS has been called from the daemon
	BOOLEAN						bIsPTSCalled;

	//pthread_t					pThreadHandle;
} READER_EXTENSION;

// this structure is specific to the IFDHANDLER of linux pcsc-lite daemon
typedef struct _SCARD_IO_HEADER {
    DWORD Protocol;
    DWORD Length;
  } SCARD_IO_HEADER, *PSCARD_IO_HEADER;

#endif
// end of file.
