/*++
 *	SCR24x USB Smart Card Reader Driver Kernel module
 *  Copyright (C) 2005-2006  TL Sudheendran(tlsudheendran@scmmicro.co.in)

 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.

 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *	SCM Microsystems  Inc., hereby disclaims all copyright interest in the
 *	library scr33x.o  written by TL Sudheendran

	Module		:	PcdmRdWr.c

	Author		:	SCM Microsystems

	Development History :

		Date		Version No.		Changes
		_______________________________________________
		02 Feb, 2004	2.0.1		Initial Versions
		30 Mar, 2004	3.0.1		Added support for suspend, resume
									under event handler.
		22 Nov, 2004	3.1.1		TTP issues 1359,1360
		25 Feb, 2005	3.2.1		Added support for Linux 2.6.x; TTP 1410, 2139
		17 May, 2005	4.1.1		Removed known limitations regarding surprise
									removal; Added LGPL License agreement



--*/

#include "includes.h"

NTSTATUS
PcdmWrite (
			PDEVICE_EXTENSION	pDeviceExtension,
			PUCHAR				pData,
			DWORD				dwDataLen
        )
/*
Routine Description :
This routine acts as the switch-case for performing the
write operation based the Pcdm Revision A or C.
*/
{
	NTSTATUS        NtStatus = STATUS_UNSUCCESSFUL;

	if(!pDeviceExtension)
	{
		DebugLog("PcdmWrite : NULL in pDeviceExtension!\n");
		return STATUS_UNSUCCESSFUL;
	}

	if(pDeviceExtension->nStop)
	{
		return NtStatus;
	}

	DebugLog ("SCR241 Pcdm Module!PcdmWrite Entry\n");
	if ( (PCDM_REV_C == pDeviceExtension->byPcdmRev) &&
			(TRUE == pDeviceExtension->bIsIrqSet ) )
	{
		DebugLog("<1> WRITE-C ENTERING\n");
		NtStatus = PcdmWriteRevC (pDeviceExtension, pData, dwDataLen);
	}
	else
	{
		DebugLog("<1> WRITE-A ENTERING\n");
		NtStatus = PcdmWriteRevA (pDeviceExtension, pData, dwDataLen);
	}

	DebugLog ("<1> SCR241 Pcdm Module::PcdmWrite Exit: %ld\n", NtStatus);
	return NtStatus;
}

NTSTATUS PcdmWriteRevA (
	DEVICE_EXTENSION	*pDeviceExtension,
	PUCHAR				pData,
	DWORD				dwDataLen
	)
/*++

Routine Description:

	This routine implements the byte write operation to transmit data
	to the device through the PCMCIA bus

	During transmission, a "START" command is sent to the firmware for
	synchronization. The BUSY bit is set and the driver waits until the
	device clears the BUSY bit. Then it places the data in the data registers
	1 to 5, and when the device clears the BUSY bit, it sends a "WRITE BYTE"
	command to the firmware. This handshake mechanism is implemented for every
	5 bytes of data that is written.

Arguments:

	pDeviceExtension	- Supplies the device extension.

	pIrp				- Supplies the I/O request packet.

Return Value:

	STATUS_SUCCESS		- Success
	STATUS_UNSUCCESSFUL - The requested operation was unsuccessful

--*/
{
	NTSTATUS			NtStatus 	= STATUS_SUCCESS;
	PPCDM_REGISTERS		IoBase;
	u_long				ulBufferLength,
						ulIdx;

	ulBufferLength		= dwDataLen;

	IoBase				= (PPCDM_REGISTERS)pDeviceExtension->IoBase;
	pDeviceExtension->tempCount	=0;

 	if (!IoBase)
	{
		DebugLog("PCDMWriteRevA : NULL in IoBase!\n");
		return STATUS_UNSUCCESSFUL;
	}

	do
	{
		NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
		if (!NT_SUCCESS (NtStatus))
		{
			break;
		}

		// Send START command to the device
		outb( CMD_START,(unsigned)&IoBase->CmdStatusReg);
		NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
		if (!NT_SUCCESS (NtStatus))
		{
			break;
		}

		for ( ulIdx = 0; ulIdx < ulBufferLength; ulIdx++ )
		{
			// Send Data to the device
			outb(pData[ulIdx++],(unsigned)&IoBase->DataReg1);
			outb(pData[ulIdx++],(unsigned)&IoBase->DataReg2);
			outb(pData[ulIdx++],(unsigned)&IoBase->DataReg3);
			outb(pData[ulIdx++],(unsigned)&IoBase->DataReg4);
			outb(pData[ulIdx],(unsigned)&IoBase->DataReg5);


			// Send WRITE command to the device
			outb(CMD_WRITE_BYTE, (unsigned)&IoBase->CmdStatusReg);

			NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
			if (!NT_SUCCESS (NtStatus))
			{
				break;
			}
			NtStatus = inb( (unsigned) &IoBase->CmdStatusReg );
			if( NtStatus & STS_ERR )
			{
					NtStatus = STATUS_UNSUCCESSFUL;
					break;
			}
		}

	} while ( FALSE ) ;

	return NtStatus;
}

NTSTATUS
PcdmWriteRevC (
        PDEVICE_EXTENSION       pDeviceExtension,
        PUCHAR                  pData,
		DWORD                   dwDataLen
        )
{
	NTSTATUS        	NtStatus 		= STATUS_SUCCESS;
	ULONG   			ulBufferLength,
						ulIdx;
	UCHAR   			Status;
	BYTE				byCharacter;
	PPCDM_REGISTERS		IoBase;

	IoBase				= (PPCDM_REGISTERS)pDeviceExtension->IoBase;
	if (!IoBase)
	{
		DebugLog("PCDMWriteRevC : NULL in IoBase!\n");
		return STATUS_UNSUCCESSFUL;
	}

	pDeviceExtension->tempCount	=0;
	ulBufferLength	= dwDataLen;

	if (0 != (ulBufferLength % 2) )
	{
			ulBufferLength++;
	}
	DebugLog("SCR241 Pcdm Module!WriteRevC Buffer Length: %ld\n", ulBufferLength);

	do
	{
		NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
		if (!NT_SUCCESS (NtStatus))
		{
			break;
		}

		// Send START command to the device
		outb( CMD_START, (unsigned) &IoBase->CmdStatusReg );

		NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
		if (!NT_SUCCESS (NtStatus))
		{
			break;
		}

		// Count in Data reg 4 & 5
		byCharacter = LOBYTE(ulBufferLength);
		outb(byCharacter, (unsigned) &IoBase->DataReg4 );
		byCharacter = HIBYTE(ulBufferLength);
		outb(byCharacter, (unsigned) &IoBase->DataReg5 );

		// Send WRITE command to the device
		outb( CMD_WRITE_BYTE, (unsigned) &IoBase->CmdStatusReg );

		// Wait for DRQ to set
		do
		{
			if ( pDeviceExtension->nStop )
			{
				DebugLog ("<1> *** Device removed! ***\n");
				break;
			}

			Status = inb( (unsigned) &IoBase->CmdStatusReg );
			//interruptible_sleep_on_timeout(&qtest, (HZ/5) );

			if ( 0xFF == Status )
			{
				// In cases where the device is unplugged during active I/O
				// all ports probably read 0xff. The firmware does not set
				// this bit by design. We make use of this information as
				// the terminal condition for this polling and flag the
				// failure, as the device has been removed
				DebugLog( "<1> Device removed \n");
				NtStatus = STATUS_UNSUCCESSFUL;
				break;
			}
			if (FALSE == (Status & STS_BSY))
			{
				if ( Status & STS_ERR )
				{
					DebugLog("<1>Error bit set\n");
					NtStatus = STATUS_UNSUCCESSFUL;
					break;
			}
				if ( (Status & STS_DRQ) )
				{
					DebugLog("<1> DRQ bit set\n");
					break;
				}
			}
		}while (TRUE); // end of wait for DRQ

		if (!NT_SUCCESS (NtStatus))
		{
			break;
		}
		// Write all the Data available
		for (ulIdx = 0; ulIdx < ulBufferLength; ulIdx++)
		{
			DebugLog(" <1> SCR241 Pcdm Module!Write Data: %x\n", pData[ulIdx]);
			outb( pData[ulIdx], (unsigned) &IoBase->DummyReg0 );

			if ( pDeviceExtension->nStop )
			{
				DebugLog ("<1> *** Device removed! ***\n");

				break;
			}

		}
	} while (FALSE);

	DebugLog("<1> SCR241 Pcdm Module!PcdmWriteRevC Exit: %ld\n",NtStatus);
	return NtStatus;
}

NTSTATUS
PcdmRead (
        PDEVICE_EXTENSION       pDeviceExtension,
        PUCHAR			pData,
	PDWORD			pdwDataLen
        )
{
	NTSTATUS        NtStatus = STATUS_UNSUCCESSFUL;

	if(!pDeviceExtension)
	{
		DebugLog("PcdmRead : NULL in pDeviceExtension!\n");
		return STATUS_UNSUCCESSFUL;
	}

	if(pDeviceExtension->nStop)
	{
		return NtStatus;
	}

	DebugLog("SCR241 Pcdm Module!PcdmRead Entry\n");

	if ((PCDM_REV_C == pDeviceExtension->byPcdmRev) &&
			(TRUE == pDeviceExtension->bIsIrqSet))
	{
			NtStatus = PcdmReadRevC (pDeviceExtension, pData, pdwDataLen);
	}
	else
	{
			NtStatus = PcdmReadRevA (pDeviceExtension, pData, pdwDataLen);
	}

	DebugLog ("PcdmRead Exit: %ld\n", NtStatus);

	return NtStatus;
}

NTSTATUS
PcdmReadRevA(
	DEVICE_EXTENSION *pDeviceExtension,
	u_char *	pData,
	PDWORD		pdwDataLen
	)
/*++

Routine Description:

	This routine implements the byte read operation to receive data
	from the device through the PCMCIA bus.
:
	During reception, the PCMCIA Client Driver sends a "READ BYTE" command
	to the firmware. The BUSY bit is set and the driver waits until the	device
	clears the BUSY bit. Then it reads the data from the data register1 to
	data register5. This handshake mechanism is implemented for every 5 bytes
	of data that is read.

Arguments:
	pDeviceExtension	- Supplies the device extension.

Return Value:

	STATUS_SUCCESS		- Success
	STATUS_UNSUCCESSFUL - The requested operation was unsuccessful

--*/
{
	NTSTATUS			NtStatus = STATUS_SUCCESS;
	PPCDM_REGISTERS		IoBase;
	u_long			ulBufferLength, ulBufferIndex;


	IoBase =(PPCDM_REGISTERS) pDeviceExtension->IoBase;

	if (!IoBase)
	{
		return STATUS_UNSUCCESSFUL;
	}

	pDeviceExtension->tempCount=0;

	// Get Response Header First.
	ulBufferLength = CCID_HEADER_SIZE;
	ulBufferIndex  = 0;

	do
	{

		NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
		if (!NT_SUCCESS (NtStatus))
		{
			DebugLog("PCDMReadRevA : NULL in IoBase!\n");
			break;
		}

		// Send READ command to the device
		outb(CMD_READ_BYTE,(unsigned) &IoBase->CmdStatusReg);

		NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
		if (!NT_SUCCESS (NtStatus))
		{
			break;
		}

		// Get the Data from the device
		pData[ulBufferIndex]	= inb( (unsigned)&IoBase->DataReg1 );
		ulBufferIndex++;
		pData[ulBufferIndex]	= inb( (unsigned)&IoBase->DataReg2 );
		ulBufferIndex++;
		pData[ulBufferIndex]	= inb((unsigned) &IoBase->DataReg3 );
		ulBufferIndex++;
		pData[ulBufferIndex]	= inb((unsigned) &IoBase->DataReg4 );
		ulBufferIndex++;
		pData[ulBufferIndex] 	= inb( (unsigned)&IoBase->DataReg5 );
		ulBufferIndex++;

		DebugLog("<1> Length is %02X,%02X \n", pData[LENGTH_LOWORD_LOBYTE_OFFSET],pData[LENGTH_LOWORD_HIBYTE_OFFSET]);
		// Check if more data is available
		ulBufferLength = MAKEWORD ( pData[LENGTH_LOWORD_LOBYTE_OFFSET],\
									pData[LENGTH_LOWORD_HIBYTE_OFFSET] );

		*pdwDataLen = (DWORD )(ulBufferLength + CCID_HEADER_SIZE) ;

		for ( ; ulBufferIndex < (ulBufferLength + CCID_HEADER_SIZE) ; ulBufferIndex++ )
		{
			if ( pDeviceExtension->nStop )
			{
				DebugLog ("<1> Device Removed!\n");
				*pdwDataLen = -1;
				break;
			}

			NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
			if (!NT_SUCCESS (NtStatus))
			{
				break;
			}

			// Send READ command to the device
			outb( CMD_READ_BYTE,(unsigned)&IoBase->CmdStatusReg);

			NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
			if (!NT_SUCCESS (NtStatus))
			{
				break;
			}

			// Get the Data from the device
			pData[ulBufferIndex]	= inb( (unsigned)&IoBase->DataReg1 );
// 			DebugLog("<1> Read-Rev-A::Data Recieved is %02X \n", pData[ulBufferIndex] );
			ulBufferIndex++;
			pData[ulBufferIndex]	= inb((unsigned) &IoBase->DataReg2 );
// 			DebugLog("<1> Read-Rev-A::Data Recieved is %02X \n", pData[ulBufferIndex] );
			ulBufferIndex++;
			pData[ulBufferIndex]	= inb((unsigned) &IoBase->DataReg3 );
// 			DebugLog("<1> Read-Rev-A::Data Recieved is %02X \n", pData[ulBufferIndex] );
			ulBufferIndex++;
			pData[ulBufferIndex]	= inb((unsigned) &IoBase->DataReg4 );
// 			DebugLog("<1> Read-Rev-A::Data Recieved is %02X \n", pData[ulBufferIndex] );
			ulBufferIndex++;
			pData[ulBufferIndex]	= inb((unsigned) &IoBase->DataReg5 );
// 			DebugLog("<1> Read-Rev-A::Data Recieved is %02X \n", pData[ulBufferIndex] );
		}

	} while( FALSE ) ;

	return NtStatus;
}

NTSTATUS
PcdmReadRevC (
	PDEVICE_EXTENSION 	pDeviceExtension,
	u_char 				*pData,
	PDWORD				pdwDataLen
        )
{
	NTSTATUS        	NtStatus 		= STATUS_SUCCESS;
	ULONG   			ulBufferLength 		= 0,
						ulBufferIndex;
	PPCDM_REGISTERS 	IoBase;

	pDeviceExtension->bIsEventSet = FALSE;

	DebugLog("SCR241 PCDM KERNEL MODULE!PcdmReadRevC Entry\n");

	IoBase = (PPCDM_REGISTERS) pDeviceExtension->IoBase;
	pDeviceExtension->tempCount=0;

	if (!IoBase)
	{
		DebugLog("PCDMReadRevC : NULL in IoBase!\n");
		return STATUS_UNSUCCESSFUL;
	}

	do
	{
		NtStatus = PcdmWaitForDeviceReady(pDeviceExtension) ;
		if (!NT_SUCCESS (NtStatus))
		{
				break;
		}
		DebugLog("Write CMD_READ\n");

		// Send READ command to the device
		outb( CMD_READ_BYTE, (unsigned) &IoBase->CmdStatusReg );
		DebugLog("Enter Wait\n");

		pDeviceExtension->bSignalEvent = FALSE;
		wait_event_interruptible( pDeviceExtension->tqsReadEventWait, (TRUE == pDeviceExtension->bIsEventSet) );

// 		Resetting the flags for the wait condition
		pDeviceExtension->bIsEventSet = FALSE;
		pDeviceExtension->bSignalEvent = FALSE;

		DebugLog("\nOut of Wait\n");

		pData[0] = inb( (unsigned) &IoBase->DataReg4 );
		pData[1] = inb( (unsigned) &IoBase->DataReg5 );

		ulBufferLength = MAKEWORD ( pData[0],pData[1] );
		*pdwDataLen = ulBufferLength;

		DebugLog ("SCR241 Pcdm Module!Buffer length: %ld\n",ulBufferLength);
		if (0 != (ulBufferLength % 2))
		{
				ulBufferLength++;
		}

		for (ulBufferIndex = 0; ulBufferIndex < ulBufferLength; ulBufferIndex++)
		{
				pData[ulBufferIndex] = inb( (unsigned) &IoBase->DummyReg0 );
				DebugLog ("SCR241 Pcdm Module!Read Data: %x\n", pData[ulBufferIndex]);

				if ( pDeviceExtension->nStop )
				{
					DebugLog ("<1> *** Device removed! ***\n");
					*pdwDataLen = -1;
					break;
				}
		}
	} while (FALSE);

	//up( &mMutex );
	DebugLog("SCR241 Pcdm Module!PcdmReadRevC Exit: %ld\n", NtStatus);
	return NtStatus;
}

NTSTATUS
PcdmWaitForDeviceReady(
	DEVICE_EXTENSION *pDeviceExtension
	)
/*++
Routine Description:
	This routine waits for the PCDM to become ready for read/write
	operations.
	This function waits for the STS_BSY line to be clear. Also, in
	case of STS_ERR being set, this flags the device as not ready
	for operation.

Arguments:
	pDeviceExtension	- Supplies the device extension.

Return Value:
	STATUS_SUCCESS		- Success
	STATUS_UNSUCCESSFUL - The requested operation was unsuccessful

--*/
{
	NTSTATUS 		NtStatus 	= STATUS_UNSUCCESSFUL;
	PPCDM_REGISTERS	IoBase		= (PPCDM_REGISTERS)pDeviceExtension->IoBase;
	u_char			Status;

	if (!IoBase)
	{
		DebugLog("PcdmWaitForDeviceReady : NULL in IoBase!\n");
		return STATUS_UNSUCCESSFUL;
	}

	do
	{
		pDeviceExtension->tempCount++;
		Status = inb ((unsigned) &IoBase->CmdStatusReg );

		if ( Status & STS_ERR )
		{
			// In cases where the device is unplugged during active I/O
			// all ports probably read 0xff. The firmware does not set
			// this bit by design. We make use of this information as
			// the terminal condition for this polling and flag the
			// failure, as the device has been removed
			DebugLog("<1> Device is Removed during WaitforReady. Cannot read/write data \n");
			break;
		}

		if ( 0 == ( Status & STS_BSY ) )
		{
			NtStatus = STATUS_SUCCESS;
			break;
		}
	} while ( TRUE );

	return NtStatus ;
}

// end of file
