/*
 *	SCR24x USB Smart Card Reader Driver Kernel module
 *  Copyright (C) 2005-2006  TL Sudheendran(tlsudheendran@scmmicro.co.in)

 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.

 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *	SCM Microsystems  Inc., hereby disclaims all copyright interest in the
 *	library scr33x.o  written by TL Sudheendran


	Module		:	ccid.h

	Author		:	SCM Microsystems

	Development History :

		Date		Version No.		Changes
		_______________________________________________
		02 Feb, 2004	2.0.1		Initial Versions
		30 Mar, 2004	3.0.1		Added support for suspend, resume
									under event handler.
		22 Nov, 2004	3.1.1		TTP issues 1359,1360
		25 Feb, 2005	3.2.1		Added support for Linux 2.6.x; TTP 1410, 2139
		17 May, 2005	4.1.1		Removed known limitations regarding surprise
									removal; Added LGPL License agreement


--*/




#ifndef _CCID_H_
#define _CCID_H_

#include "includes.h"
//
// Defines Regarding Offsets in the Structure with Respect to Bytes.
//

#define MESSAGE_TYPE_OFFSET					0
#define LENGTH_LOWORD_LOBYTE_OFFSET			1
#define LENGTH_LOWORD_HIBYTE_OFFSET			2
#define LENGTH_HIWORD_LOBYTE_OFFSET			3
#define LENGTH_HIWORD_HIBYTE_OFFSET			4
#define SLOT_OFFSET							5
#define SEQUENCE_OFFSET						6
#define STATUS_REG_OFFSET					7
#define ERROR_REG_OFFSET					8
#define RFU_BYTE_OFFSET						9
#define ESCAPE_COMMAND_OFFSET				10
#define ESCAPE_DATA_OFFSET					11
#define DATA_OFFSET							10
#define MAX_CCID_OUTBUFF_SIZE				512
#define MAX_CCID_DATA_SIZE					264
#define MAX_INTELLIGENT_SIZE				260


//
//  Command Message Types
//
#define PC_TO_RDR_ICCPOWERON				0x62
#define PC_TO_RDR_ICCPOWEROFF				0x63
#define PC_TO_RDR_GETSLOTSTATUS				0x65
#define PC_TO_RDR_XFRBLOCK					0x6F
#define PC_TO_RDR_GETPARAMETERS				0x6C
#define PC_TO_RDR_RESETPARAMETERS			0x6D
#define PC_TO_RDR_SETPARAMETERS				0x61
#define PC_TO_RDR_ESCAPE					0x6B
#define PC_TO_RDR_ICCCLOCK					0x6E
#define PC_TO_RDR_T0APDU					0x6A
#define PC_TO_RDR_SECURE					0x69
#define PC_TO_RDR_MECHANICAL				0x71
#define PC_TO_RDR_ABORT								0x72
#define PC_TO_RDR_SETDATARATEANDCLOCKFREQUENCY		0x73

//
//  Interrupt endpoint CCID messages definitions
//
#define RDR_TO_PC_NOTIFYSLOTCHANGE			0x50
#define RDR_TO_PC_HARDWAREERROR				0x51
#define RDR_TO_PC_PINENTRYCHANGE			0xA0

//
// Response Message Types
//
#define RDR_TO_PC_DATABLOCK					0X80
#define RDR_TO_PC_SLOTSTATUS				0X81
#define RDR_TO_PC_PARAMETERS				0X82
#define RDR_TO_PC_ESCAPE					0X83
#define RDR_TO_PC_DATARATEANDCLOCKFREQUENCY	0X84

//
// Slot ID
//
#define MASTER_SLOT_ID						0

//
// Escape Commands Code used in abyMessageSpecific[2]
//
#define READER_GETINFO						0x00
#define READER_SETMODE						0x01
#define READER_GETMODE						0x02
#define READER_UNDEFINED1					0x03
#define READER_SET_DBG_INFO					0x04
#define READER_EMV_LOOPBACK					0x05
#define READER_EMV_SINGLEMODE				0x06
#define READER_EMV_TIMERMODE				0x07
#define READER_APDU_TRANSFER				0x08
#define READER_DFU_DETACH					0x09
#define READER_SWITCH_SPEED					0x0A
#define READER_SWITCH_PROTOCOL				0x0C
#define READER_GETCHIPREV               	0x13
#define READER_SETTRANSPORT             	0x14


//
// Defines Regarding the class specific commands.
//
#define CTRL_ABORT							1

//
// Defines Regarding the Interface details
//

// Interface								Value
#define CCID_HEADER_SIZE					0x0A

// CCID Status and Error code details

//
// For bmCommandStatus bits
//
#define IFD_NO_ERROR						0x00
#define IFD_CMD_FAILED						0x40
#define IFD_TIME_EXTN_REQUESTED				0x80

//
// For bmICCStatus bits for IFD
//
#define CCID_ICC_ACTIVE						0x00
#define CCID_ICC_PRESENT					0x01
#define CCID_ICC_ABSENT						0x02

//
// Card Status Definitions
//
#define READER_STS_CARD_ABSENT      		0x00
#define READER_STS_CARD_PRESENT				0x01

#pragma pack(1)

//
// GetInfo Structure.
//
typedef struct __SCARD_READER_GETINFO_PARAMS
{
	BYTE			byStatus;
	BYTE			byError;
	BYTE			byMajorVersion;
	BYTE			byMinorVersion;
	BYTE			bySupportedModes;
	WORD			wSupportedProtocols;
	WORD			winputDevice;
	BYTE			byPersonality;
	DWORD			dwSerialNumber;
	BYTE			byMaxSlots;
}SCARD_READER_GETINFO_PARAMS,*PSCARD_READER_GETINFO_PARAMS;

 //
// GetMode Structure
//
typedef struct __SCARD_READER_GETMODE_PARAMS
{
	BYTE			byStatus;
	BYTE			byError;
	BYTE			byMode;
}SCARD_READER_GETMODE_PARAMS,*PSCARD_READER_GETMODE_PARAMS;

//
// GetCardStatus Structure
//

typedef struct __SCARD_READER_GETCARDSTATUS_PARAMS
{
	BYTE			byStatus;
	BYTE			byError;
	BYTE			byCardStatus;
}SCARD_READER_GETCARDSTATUS_PARAMS,*PSCARD_READER_GETCARDSTATUS_PARAMS;


//
// Ccid Header - 10 bytes
//
typedef struct __SCARD_CCID_HEADER
{
	BYTE			byMessageType;
	DWORD 			dwLength;
	BYTE			bySlot;
	BYTE			bySequence;
	BYTE			abyMessageSpecific[3];

}SCARD_CCID_HEADER, *PSCARD_CCID_HEADER;


//
// CCID Message Structure
//
typedef struct __SCARD_CCID_COMMAND_PARAM
{
	SCARD_CCID_HEADER	Header;
	BYTE			abyData[260];
}SCARD_CCID_COMMAND_PARAM,*PSCARD_CCID_COMMAND_PARAM;

//
// 5. Structure For Class Commands.
//
typedef struct __CLASS_COMMAND
{
	BYTE 			bymRequestType;
	BYTE 			byRequest;
	WORD 			wValue;
	WORD 			wIndex;
	WORD 			wLength;

}CLASS_COMMAND,*PCLASS_COMMAND;

#endif


// end of file.
