/*
 *	SCR24x USB Smart Card Reader Driver Kernel module
 *  Copyright (C) 2005-2006  TL Sudheendran(tlsudheendran@scmmicro.co.in)

 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.

 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *	SCM Microsystems  Inc., hereby disclaims all copyright interest in the
 *	library scr33x.o  written by TL Sudheendran


	Module		:	def.h

	Author		:	SCM Microsystems

	Development History :

		Date		Version No.		Changes
		_______________________________________________
		02 Feb, 2004	2.0.1		Initial Versions
		30 Mar, 2004	3.0.1		Added support for suspend, resume
									under event handler.
		22 Nov, 2004	3.1.1		TTP issues 1359,1360
		25 Feb, 2005	3.2.1		Added support for Linux 2.6.x; TTP 1410, 2139
		17 May, 2005	4.1.1		Removed known limitations regarding surprise
									removal; Added LGPL License agreement


--*/




#ifndef _DEF_H_
#define _DEF_H_

// current driver version
#define PCSC_MAJOR_VERSION      0x03
#define PCSC_MINOR_VERSION      0x01

#define RtlZeroMemory( pucBuffer , nSize) memset( (pucBuffer),0,nSize)
#define RtlCopyMemory( pucBuf1, pucBuf2, nSize ) memcpy( (pucBuf1), (pucBuf2), (nSize) )
#define ExFreePool( pucBuffer ) free( (pucBuffer))

#ifndef NT_SUCCESS
#define NT_SUCCESS(Status) ((0==(Status))?TRUE:FALSE)
#endif

#ifndef TRUE
  #define TRUE  							1
#endif

#ifndef FALSE
  #define FALSE 							0
#endif

#ifndef NULL
  #define NULL                     			0x0000
#endif

// IOCTL Commands
#define CMD_CARD_STATUS						11
#define CMD_PCCARD_DOWNLOAD_END				12
#define CMD_WRITE_DATA						13
#define CMD_READ_DATA						14
#define CMD_POLLED_MODE						15
#define CMD_PCCARD_ERASE					16
#define CMD_SET_TRANSPORT					17
#define CMD_PCCARD_DOWNLOAD_START			18
#define CMD_PCCARD_DOWNLOAD_DATA			19
#define CMD_GET_IRQSTATUS					81
#define CMD_SET_REVISION					82
#define CMD_BOOTROM_STATUS					83

//CMD_GENERIC_IOCTLs

#define STATUS_SUCCESS						0x0000
#define STATUS_UNSUCCESSFUL					0x0001
#define STATUS_DEVICE_PROTOCOL_ERROR		0x0002
#define STATUS_IO_TIMEOUT					0x0003
#define STATUS_BUFFER_TOO_SMALL       		0x0004
#define STATUS_INSUFFICIENT_RESOURCES 		0x0005
#define STATUS_UNRECOGNIZED_MEDIA			0x0006
#define STATUS_INVALID_PARAMETER			0x0007
#define STATUS_INVALID_DEVICE_STATE     	0x0008

#define MAKEWORD(a,b) ((u_short)(((u_char)(a) & 0xff) | (u_short)(((u_char)(b) & 0xff) << 8)))

#define DEBUG_TRACE             			((ULONG) 0x00000010)

#define MAXIMUM_ATTR_STRING_LENGTH 			32
#define MAX_BUF_LEN							2500 // updated 600 from ricoh
// Number of RFU bytes
#define MAX_RFU_BYTES						3

#define SCARD_READER_TYPE_SERIAL    		0x01
#define SCARD_READER_TYPE_PARALELL  		0x02
#define SCARD_READER_TYPE_KEYBOARD  		0x04
#define SCARD_READER_TYPE_SCSI      		0x08
#define SCARD_READER_TYPE_IDE       		0x10
#define SCARD_READER_TYPE_USB       		0x20
#define SCARD_READER_TYPE_PCMCIA    		0x40
#define SCARD_READER_TYPE_VENDOR    		0xF0

#define SCARD_PROTOCOL_UNDEFINED    		0x00000000  // no active protocol.

// This is the mask of ISO defined transmission protocos
#define SCARD_PROTOCOL_Tx           		SCARD_PROTOCOL_T0 | SCARD_PROTOCOL_T1

// Use the default transmission parameters / card clock freq.
#define SCARD_PROTOCOL_DEFAULT      		0x80000000

// Use optimal transmission parameters / card clock freq.
// Since using the optimal parameters is the default case no bit is defined to be 1

#define SCARD_PROTOCOL_OPTIMAL      		0x00000000

// Ioctl parameters 1 for IOCTL_SMARTCARD_POWER
#define SCARD_POWER_DOWN 					10 // Power down the card.
#define SCARD_WARM_RESET 					12 // Force a reset on the card.
#define SCARD_COLD_RESET 					13 // Cycle power and reset the card.

// defines used in smcfns.c

#define INS_VERIFY							0x20
#define INS_MODIFY							0x24

#define INS_ENABLE_VR						0x28
#define INS_DISABLE_VR						0x26
#define INS_RESET_RC						0x2C
#define INS_UNBLOCK_APP						0x18

#define INS_T0_FIELD						0x01
#define INS_T1_FIELD						0x04

#define T1_NAD								0x00
#define IFS_REQUEST							0xC1
#define IFS_REQUEST_LEN						0x01

#define R_BLOCK_VERIFY						0x80
#define T1_EDC_ERROR						0x81
#define T1_OTHER_ERROR						0x82

#define T1_MAX_IFSC							0x3C
#define T1_MAX_IFSD							0xFC

// IFD Handler version number enummerations
#define IFD_HVERSION_1_0               		0x00010000
#define IFD_HVERSION_2_0               		0x00020000

// List of defines available to ifdhandler

#define IFD_POWER_UP                    	500
#define IFD_POWER_DOWN                  	501
#define IFD_RESET                       	502

#define IFD_NEGOTIATE_PTS1              	1
#define IFD_NEGOTIATE_PTS2              	2
#define IFD_NEGOTIATE_PTS3              	4

#define IFD_SUCCESS                     	0
#define IFD_ERROR_TAG                   	600
#define IFD_ERROR_SET_FAILURE           	601
#define IFD_ERROR_VALUE_READ_ONLY       	602
#define IFD_ERROR_PTS_FAILURE           	605
#define IFD_ERROR_NOT_SUPPORTED         	606
#define IFD_PROTOCOL_NOT_SUPPORTED      	607
#define IFD_ERROR_POWER_ACTION          	608
#define IFD_ERROR_SWALLOW               	609
#define IFD_ERROR_EJECT                 	610
#define IFD_ERROR_CONFISCATE            	611
#define IFD_COMMUNICATION_ERROR         	612
#define IFD_RESPONSE_TIMEOUT            	613
#define IFD_NOT_SUPPORTED               	614
#define IFD_ICC_PRESENT                 	615
#define IFD_ICC_NOT_PRESENT             	616


#define IOCTL_SMARTCARD_POWER				0x310004
#define IOCTL_SMARTCARD_GET_ATTRIBUTE		0x310008
#define IOCTL_SMARTCARD_SET_ATTRIBUTE		0x31000C
#define IOCTL_SMARTCARD_CONFISCATE			0x310010
#define IOCTL_SMARTCARD_TRANSMIT			0x310014
#define IOCTL_SMARTCARD_EJECT				0x310018
#define IOCTL_SMARTCARD_SWALLOW				0x31001C
#define IOCTL_SMARTCARD_IS_PRESENT			0x310028
#define IOCTL_SMARTCARD_IS_ABSENT			0x31002C
#define IOCTL_SMARTCARD_SET_PROTOCOL		0x310030
#define IOCTL_SMARTCARD_GET_STATE			0x310038
#define IOCTL_SMARTCARD_GET_LAST_ERROR		0x31003C
#define IOCTL_SMARTCARD_CANCEL_BLOCKING		0x310040

#define MAKELONG(a, b)      				((LONG)(((WORD)((DWORD_PTR)(a) & \
											0xffff)) | \
											((DWORD)((WORD)((DWORD_PTR)(b) & \
											0xffff))) << 16))

#define HIWORD(value)           ((WORD)((value) >> 16))
#define LOWORD(value)           (WORD)(value)
#define HIBYTE(value)           ((BYTE)((value) >> 8))
#define LOBYTE(value)           ((BYTE)value)


#ifndef BYTE
	typedef unsigned char BYTE;
#endif
#ifndef DWORD
	typedef unsigned long DWORD;
#endif
#ifndef UCHAR
	typedef unsigned char UCHAR;
#endif

typedef unsigned long 	ULONG;
typedef unsigned short 	WORD;
typedef long		LONG;
typedef unsigned short 	USHORT;
typedef short           BOOLEAN;
typedef short		BOOL;
typedef BOOLEAN         bool;
typedef UCHAR           *PBYTE;
typedef char*           STR;
typedef ULONG           NTSTATUS;
typedef ULONG           HPORT;
typedef ULONG           PVMMDDB;
typedef ULONG           DEVNODE;
typedef unsigned long	*PDWORD;
typedef unsigned short  *PWORD;
typedef void*           PVOID;
typedef ULONG           VMM_SEMAPHORE;
typedef ULONG           HTIMEOUT;
typedef ULONG           HANDLE;
typedef void            VOID;
typedef ULONG           DWORD_PTR;
typedef BYTE 		TIMER_HANDLE;
typedef unsigned char  *PUCHAR;


#define CTL_CODE( DeviceType, Function, Method, Access ) (                 \
    ((DeviceType) << 16) | ((Access) << 14) | ((Function) << 2) | (Method) \
)

// callback handler defines
#define CALLBACK_TRANSMIT	0x01

#if DBG || DEBUG
#define DEBUG_IOCTL     ((ULONG) 0x00000001)
#define DEBUG_ATR       ((ULONG) 0x00000002)
#define DEBUG_PROTOCOL  ((ULONG) 0x00000004)
#define DEBUG_DRIVER    ((ULONG) 0x00000008)
#define DEBUG_TRACE     ((ULONG) 0x00000010)
#define DEBUG_ERROR     ((ULONG) 0x00000020)
#define DEBUG_INFO      DEBUG_ERROR
#define DEBUG_T1_TEST   ((ULONG) 0x40000000)
#define DEBUG_BREAK     ((ULONG) 0x80000000)
#define DEBUG_ALL       ((ULONG) 0x0000FFFF)
#endif

//
// Indexes to the callback functions of the ReaderFunction array
// in the SmartcardExtension
//
#define RDF_CARD_POWER 		0
#define RDF_TRANSMIT		1
#define RDF_CARD_EJECT		2
#define RDF_READER_SWALLOW 	3
#define RDF_CARD_TRACKING	4
#define RDF_SET_PROTOCOL	5
#define RDF_DEBUG_LEVEL		6
#define RDF_CARD_CONFISCATE 7
#define RDF_IOCTL_VENDOR    8
#define RDF_ATR_PARSE       9

//
// Minimum buffer size for request and reply buffer
//
#define MIN_BUFFER_SIZE 4096
//
#define MAXIMUM_ATTR_STRING_LENGTH 32

#define MAX_SERIAL_DATA_SIZE  (64*1024)

typedef UCHAR TCHAR;
typedef UCHAR *PTCHAR;








#endif
