/*
 *	SCR24x USB Smart Card Reader Driver Kernel module
 *  Copyright (C) 2005-2006  TL Sudheendran(tlsudheendran@scmmicro.co.in)

 *   This library is free software; you can redistribute it and/or
 *   modify it under the terms of the GNU Lesser General Public
 *   License as published by the Free Software Foundation; either
 *   version 2.1 of the License, or (at your option) any later version.

 *   This library is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *   Lesser General Public License for more details.

 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


 *	SCM Microsystems  Inc., hereby disclaims all copyright interest in the
 *	library scr33x.o  written by TL Sudheendran


	Module		:	ifd.c

	Author		:	SCM Microsystems

	Development History :

		Date		Version No.		Changes
		_______________________________________________
		02 Feb, 2004	2.0.1		Initial Versions
		30 Mar, 2004	3.0.1		Added support for suspend, resume
									under event handler.
		22 Nov, 2004	3.1.1		TTP issues 1359,1360
		25 Feb, 2005	3.2.1		Added support for Linux 2.6.x; TTP 1410, 2139
		17 May, 2005	4.1.1		Removed known limitations regarding surprise
									removal; Added LGPL License agreement


--*/





#include "includes.h"
#include "protos.h"
NTSTATUS
IFDPCCardDownloadStart (
	PDEVICE_EXTENSION pDeviceExtension,
	PUCHAR pBuffer
    )
{
	NTSTATUS			NTStatus 	= STATUS_SUCCESS;

	//Kill the Card Tracking Thread if any.
	if( FALSE == pDeviceExtension->bBootRomPresent )
	{
		if (FALSE == pDeviceExtension->bIsIrqSet)
		{
			send_sig(SIGTERM, pDeviceExtension->pDE_thread, 1);
		}
	}

	wake_up_interruptible ( &pDeviceExtension->tqsEventWait );
	pDeviceExtension->bDwnComplete = FALSE;
	return NTStatus;
}

NTSTATUS
IFDPerformPCCardDownLoad(
    PDEVICE_EXTENSION pDeviceExtension,
    PUCHAR pBuffer
    )
{
	NTSTATUS		NtStatus 	= STATUS_SUCCESS;
	PPCDM_REGISTERS		IoBase;
	static ULONG		ulLength 	= 0;
	BYTE			Status;
	USHORT			Value;
	int			nI;
	unsigned long		lFlags		= 0;

	DebugLog( "DFU::Disabling Interrupt . . . \n" );

	save_flags ( lFlags );
	cli ();

	interruptible_sleep_on_timeout(&pDeviceExtension->wait,( HZ*10) );

	IoBase	= (PPCDM_REGISTERS) pDeviceExtension->IoBase;

	DebugLog("IFDPerformPCCardDownLoad - Preparing for DFU \n");

	// Build and send Download Command
	outb(0x40, (unsigned) &IoBase->DummyReg6);
	outb(0x07, (unsigned) &IoBase->DataReg1);
	outb(0x80, (unsigned) &IoBase->DataReg2);
	outb(0x00, (unsigned) &IoBase->DataReg3);
	outb(0x00, (unsigned) &IoBase->DataReg4);
	outb(0x00, (unsigned) &IoBase->DataReg5);
	outb(0x92, (unsigned) &IoBase->CmdStatusReg);

	// Wait for BSY to clear and DRQ to be asserted
	// and Write Data 512 Bytes at a time.
	do
	{
		do
		{
			Status = inb ((unsigned) &IoBase->CmdStatusReg );

			if ( 0xFF == Status )
			{
				// In cases where the device is unplugged during active I/O
				// all ports probably read 0xff. The firmware does not set
				// this bit by design. We make use of this information as
				// the terminal condition for this polling and flag the
				// failure, as the device has been removed
				DebugLog("Device removed \n");
				NtStatus = STATUS_UNSUCCESSFUL;
				break;
			}

			if (FALSE == (Status & STS_BSY))
			{
				if ( Status & STS_ERR )
				{
					DebugLog("Error bit set\n");
					NtStatus = STATUS_UNSUCCESSFUL;
					break;
				}

				if( ulLength >= 65535)
				{
					// terminal condition reached! All sectors have been
					// programmed successfully
					DebugLog("Terminal condition reached\n");
					break;
				}

				if (Status & STS_DRQ)
				{
					DebugLog("DRQ bit set. ");
					break;
				}

			}

		}while (TRUE);

		if( NtStatus == STATUS_UNSUCCESSFUL )
		{
			DebugLog("Status Unsuccessful\n");
			break;
		}
		else
		{
			if( ulLength >= 65535)
			{
				DebugLog("Status Successful\n");
				break;
			}
		}

		if (TRUE == pDeviceExtension->b256WordDfu)
		{
			for(nI = 0; nI<256; nI++)
			{
				//DebugLog("%ld::",ulLength);
				//DebugLog("%02X,%02X>|", pBuffer[ulLength],pBuffer[ulLength+1] );
				Value = MAKEWORD( pBuffer[ulLength],pBuffer[ulLength+1] );
				outw(Value, (unsigned) &IoBase->DummyReg0 );
				ulLength++;
				ulLength++;
			}

			DebugLog("ulLength: %ld\n", ulLength);
		}
		else
		{
			DebugLog( "<=======512 bytes of data:: ulLength: %ld=========> \n", ulLength);
			for ( nI = 0; nI < 512; nI++)
			{
				outb(pBuffer[ulLength], (unsigned) &IoBase->DummyReg0 );
				ulLength++;
			}

		}

	}while( TRUE );

	DebugLog("IFDPerformPCCardDownLoad - Completed giving data to BOOTROM \n");

	if( NT_SUCCESS(NtStatus) )
	{
		pDeviceExtension->bDwnComplete = TRUE;
	}
	// Now give a PCMCIA Reset to jump to Functional Mode.
	if(NT_SUCCESS(NtStatus))
	{
		DebugLog("Data written successfully. moving to functional mode \n");
		outb(0x04, (unsigned) &IoBase->DeviceControlReg);
		DebugLog("IFDPerformPCCardDownLoad - Entering into delay \n");
		interruptible_sleep_on_timeout(&pDeviceExtension->wait,100);
		outb(0x00, (unsigned) &IoBase->DeviceControlReg);
		DebugLog ("IFDPerformPCCardDownLoad - Coming out from delay \n");
	}
	DebugLog("IFDPerformPCCardDownLoad Exit \n");

	restore_flags ( lFlags );

	// Firmware download complete. Hence, restart
	// card status thread if no IRQ is present
	if (FALSE == pDeviceExtension->bIsIrqSet)
	{
		pDeviceExtension->DE_thread_pid = kernel_thread(PcdmCardStatusThread, pDeviceExtension, CLONE_KERNEL);
	}

	pDeviceExtension->bBootRomPresent = FALSE;
	pDeviceExtension->bDwnComplete = TRUE;

	if( PCDM_REV_C == pDeviceExtension->byPcdmRev )
	{
		pDeviceExtension->byPcdmRev = PCDM_REV_A;
		NtStatus = IFDSetTransportMechanism( pDeviceExtension );
		if( 0 == NtStatus )
		{
			pDeviceExtension->byPcdmRev = PCDM_REV_C;
		}
		else
		{
			DebugLog("<1>Failure in Setting transport Mechanism \n" );
		}
	}

	return NtStatus;
}

void IFDPerformPCCardErase(
	PDEVICE_EXTENSION pDeviceExtension )
{
	PPCDM_REGISTERS		IoBase;
	IoBase = (PPCDM_REGISTERS) pDeviceExtension->IoBase;
// 	outb( 0x93, (unsigned) &IoBase->CmdStatusReg );
}

NTSTATUS
IFDGetInterfaceDetails (
	PDEVICE_EXTENSION	pDeviceExtension
	)
{
	PSCARD_CCID_COMMAND_PARAM	pCCIDParam 	= NULL;
	PUCHAR				pResBuffer 	= NULL;
	PUCHAR				pBuffer		= NULL;
	PIOCTL_DATA			pIoctlData	= NULL;
	BYTE				byIndex 	= 0;
	NTSTATUS 			NTStatus	= STATUS_UNSUCCESSFUL;
	PDWORD				pdwDataLen	= NULL;

	do
	{
		pdwDataLen = (PDWORD) kmalloc( (sizeof(DWORD) * 1), GFP_KERNEL );
		if( NULL == pdwDataLen )
		{
			NTStatus = STATUS_INSUFFICIENT_RESOURCES;
			break;
		}

		pBuffer = (PUCHAR) kmalloc(sizeof(SCARD_CCID_COMMAND_PARAM), GFP_KERNEL);
		if (NULL == pBuffer)
		{
			DebugLog  ("CCID IFD Module!Allocation Error\n");
			break;
		}

		pIoctlData = (PIOCTL_DATA) kmalloc(sizeof( IOCTL_DATA), GFP_KERNEL);
		if(NULL == pIoctlData )
		{
			DebugLog ("CCID IFD Module!Allocation Error\n");
			break;
		}

		memset(pIoctlData, 0, sizeof(IOCTL_DATA));

		pResBuffer = (u_char *) kmalloc ((sizeof(u_char) * MAX_CCID_OUTBUFF_SIZE), GFP_KERNEL);
		if (NULL == pResBuffer)
		{
			NTStatus = STATUS_INSUFFICIENT_RESOURCES;
			break;
		}
		memset(pResBuffer, 0, MAX_CCID_OUTBUFF_SIZE);

		pCCIDParam = (PSCARD_CCID_COMMAND_PARAM)pBuffer;

		memset (pCCIDParam, 0, sizeof (SCARD_CCID_COMMAND_PARAM));


		// Use READER_GETINFO to get the Firmware Major and Minor Versions
		// from the device
		// Build Command header
		pCCIDParam->Header.byMessageType	= PC_TO_RDR_ESCAPE;
		pCCIDParam->Header.dwLength			= 1;
		pCCIDParam->Header.bySlot			= MASTER_SLOT_ID;
		pCCIDParam->Header.bySequence		= 1;

		// RFU bytes
		for (byIndex = 0; byIndex < MAX_RFU_BYTES; byIndex++)
		{
			pCCIDParam->Header.abyMessageSpecific [byIndex] = 0x00;
		}

		pCCIDParam->abyData[0] = READER_GETINFO;

		// Data Length + Command Header Length
		pDeviceExtension->ulDataLength = \
				pCCIDParam->Header.dwLength + CCID_HEADER_SIZE;

		// We send the IOCTL to write and read to the kernel module for transaction
		memcpy( pIoctlData->ucRxBuffer , pCCIDParam, pDeviceExtension->ulDataLength );
		pIoctlData->dwRxLength = pDeviceExtension->ulDataLength;

		// Write the command
		NTStatus = PcdmWrite (pDeviceExtension, (u_char *)pCCIDParam, pDeviceExtension->ulDataLength);

		DebugLog ("<1> Error value [PcdmWrite] : %ld\n", NTStatus);

		if (!NT_SUCCESS (NTStatus))
		{
			break;
		}

		// Read the response
		NTStatus = PcdmRead ( pDeviceExtension, pResBuffer, pdwDataLen );

		DebugLog ("<1> Error value [PcdmRead] : %ld\n", NTStatus);

		if (!NT_SUCCESS (NTStatus))
		{
			break;
		}

		DebugLog( "Major Version is %02X. Minor Version is %02X \n", pResBuffer[DATA_OFFSET], pResBuffer[DATA_OFFSET + 1] );
		// Check the Response Type and Sequence Number
		if ((RDR_TO_PC_ESCAPE == pResBuffer[MESSAGE_TYPE_OFFSET]) && \
			(pDeviceExtension->bySequence == pResBuffer[SEQUENCE_OFFSET]))
		{
			if( ( 0x00 == pResBuffer[DATA_OFFSET]) || ( 0x00 == pResBuffer[DATA_OFFSET + 1]) )
			{
				DebugLog("IFDInterfaceDetails:: Device is in BootRom Mode \n" );
				pDeviceExtension->bBootRomPresent = TRUE;
				NTStatus = STATUS_SUCCESS;
			}
			else
			{
				DebugLog( "IFDInterfaceDetails:: Device Fw Version is %02X.%02X \n" ,
				pResBuffer[DATA_OFFSET], pResBuffer[DATA_OFFSET + 1] );
				pDeviceExtension->bBootRomPresent = FALSE;
				NTStatus = STATUS_SUCCESS;
			}
		}
		else
		{
			if( ( 0x00 == pResBuffer[DATA_OFFSET]) || ( 0x00 == pResBuffer[DATA_OFFSET + 1]) )
			{
				DebugLog("IFDInterfaceDetails:: Device is in BootRom Mode \n" );
				pDeviceExtension->bBootRomPresent = TRUE;
				NTStatus = STATUS_SUCCESS;
			}
		}

		pDeviceExtension->bySequence++;
	}while(FALSE);

	if( pdwDataLen)
	{
		kfree( pdwDataLen );
		pdwDataLen = NULL;
	}

	if( pIoctlData )
	{
		kfree( pIoctlData );
		pIoctlData = NULL;
	}

	if (pBuffer)
	{
		kfree (pBuffer);
		pBuffer = NULL;
	}

	if (pResBuffer != NULL)
	{
		kfree (pResBuffer);
		pResBuffer = NULL;
	}

	return NTStatus;
}

NTSTATUS IFDSetTransportMechanism
	(PDEVICE_EXTENSION pDeviceExtension )
{
	PSCARD_CCID_COMMAND_PARAM		pCCIDParam 	= NULL;
	PUCHAR							pResBuffer 	= NULL;
	NTSTATUS                        Status 		= STATUS_SUCCESS;
	BYTE                            byIndex 	= 0;
	PIOCTL_DATA						pIoctlData	= NULL;
	NTSTATUS						NTStatus	= STATUS_UNSUCCESSFUL;
	BYTE							byRev;

	byRev = PCDM_REV_C;

	DebugLog("SCR241 IFD KERNEL Module!IFDSetTransportMechanism Entry\n");

	do
	{
		pIoctlData = (PIOCTL_DATA)
			kmalloc( sizeof( IOCTL_DATA ) , GFP_KERNEL );
		if (NULL == pIoctlData)
		{
				Status = STATUS_INSUFFICIENT_RESOURCES;
				break;
		}
		pResBuffer = (PUCHAR) kmalloc( (sizeof(UCHAR) * MAX_CCID_OUTBUFF_SIZE) , GFP_KERNEL );
		if (NULL == pResBuffer)
		{
				Status = STATUS_INSUFFICIENT_RESOURCES;
				break;
		}
		memset (pIoctlData, 0, sizeof (IOCTL_DATA));
		memset (pResBuffer, 0, MAX_CCID_OUTBUFF_SIZE);

		pCCIDParam = (PSCARD_CCID_COMMAND_PARAM) pResBuffer;

		// Build Command header
		pCCIDParam->Header.byMessageType	= PC_TO_RDR_ESCAPE;
		pCCIDParam->Header.dwLength         = 2;
		pCCIDParam->Header.bySlot           = MASTER_SLOT_ID;
		pCCIDParam->Header.bySequence   	= pDeviceExtension->bySequence;

		// RFU bytes
		for (byIndex = 0; byIndex < MAX_RFU_BYTES; byIndex++)
		{
				pCCIDParam->Header.abyMessageSpecific [byIndex] = 0x00;
		}
		// Escape function
		pCCIDParam->abyData [0] = READER_SETTRANSPORT;
		// Data byte
		pCCIDParam->abyData [1] = byRev;

		// Data Length + Command Header Length
		pDeviceExtension->ulDataLength = \
						pCCIDParam->Header.dwLength + CCID_HEADER_SIZE;

		// We send the IOCTL to write and read to the kernel module for transaction
		memcpy( pIoctlData->ucRxBuffer , pResBuffer, pDeviceExtension->ulDataLength );
		pIoctlData->dwRxLength = pDeviceExtension->ulDataLength;

		// Write the command
		NTStatus = PcdmWrite (
							pDeviceExtension,
							(u_char *)pCCIDParam,
							pDeviceExtension->ulDataLength
							);

		Status = NTStatus;

		pDeviceExtension->bySequence++;
	} while ( FALSE );

	if (pIoctlData != NULL)
	{
			kfree (pIoctlData);
			pIoctlData = NULL;
	}

	if (pResBuffer != NULL)
	{
			kfree (pResBuffer);
			pResBuffer = NULL;
	}
	DebugLog("<1> *** SCR241 IFD KERNEL MODULE!IFDSetTransportMechanism Exit: %ld ***\n", Status);
	return Status;
}

// end of file

